package org.eclipse.jdt.internal.compiler.flow;

package .tmp;

import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.TryFlowContext;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;

public class FinallyFlowContext extends TryFlowContext {
  Reference[] finalAssignments;
  
  VariableBinding[] finalVariables;
  
  int assignCount;
  
  LocalVariableBinding[] nullLocals;
  
  ASTNode[] nullReferences;
  
  int[] nullCheckTypes;
  
  int nullCount;
  
  public FlowContext tryContext;
  
  public FinallyFlowContext(FlowContext parent, ASTNode associatedNode, ExceptionHandlingFlowContext tryContext) {
    super(parent, associatedNode);
    this.tryContext = (FlowContext)tryContext;
  }
  
  public void complainOnDeferredChecks(FlowInfo flowInfo, BlockScope scope) {
    int i;
    for (i = 0; i < this.assignCount; i++) {
      VariableBinding variable = this.finalVariables[i];
      if (variable != null) {
        boolean complained = false;
        if (variable instanceof FieldBinding) {
          if (flowInfo.isPotentiallyAssigned((FieldBinding)variable)) {
            complained = true;
            scope.problemReporter().duplicateInitializationOfBlankFinalField((FieldBinding)variable, this.finalAssignments[i]);
          } 
        } else if (flowInfo.isPotentiallyAssigned((LocalVariableBinding)variable)) {
          variable.tagBits &= 0xFFFFFFFFFFFFF7FFL;
          if (variable.isFinal()) {
            complained = true;
            scope.problemReporter().duplicateInitializationOfFinalLocal(
                (LocalVariableBinding)variable, 
                (ASTNode)this.finalAssignments[i]);
          } 
        } 
        if (complained) {
          FlowContext currentContext = getLocalParent();
          while (currentContext != null) {
            currentContext.removeFinalAssignmentIfAny(this.finalAssignments[i]);
            currentContext = currentContext.getLocalParent();
          } 
        } 
      } 
    } 
    if ((this.tagBits & 0x1) != 0) {
      for (i = 0; i < this.nullCount; i++) {
        int nullStatus;
        ASTNode location = this.nullReferences[i];
        switch (this.nullCheckTypes[i] & 0xFFFF0FFF) {
          case 128:
            nullStatus = flowInfo.nullStatus(this.nullLocals[i]);
            if (nullStatus != 4)
              this.parent.recordNullityMismatch(scope, (Expression)location, 
                  this.providedExpectedTypes[i][0], this.providedExpectedTypes[i][1], flowInfo, nullStatus, null); 
            break;
          case 16:
            checkUnboxing((Scope)scope, (Expression)location, flowInfo);
            break;
          default:
            this.parent.recordUsingNullReference((Scope)scope, this.nullLocals[i], 
                this.nullReferences[i], this.nullCheckTypes[i], flowInfo);
            break;
        } 
      } 
    } else {
      for (i = 0; i < this.nullCount; i++) {
        Expression expression;
        int nullStatus;
        ASTNode location = this.nullReferences[i];
        LocalVariableBinding local = this.nullLocals[i];
        switch (this.nullCheckTypes[i] & 0xFFFF0FFF) {
          case 256:
          case 512:
            if (flowInfo.isDefinitelyNonNull(local)) {
              if ((this.nullCheckTypes[i] & 0xFFFF0FFF) == 512) {
                if ((this.nullCheckTypes[i] & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNonNull(local, location); 
                break;
              } 
              scope.problemReporter().localVariableNonNullComparedToNull(local, location);
              break;
            } 
          case 257:
          case 513:
          case 769:
          case 1025:
            expression = (Expression)location;
            if (flowInfo.isDefinitelyNull(local)) {
              switch (this.nullCheckTypes[i] & 0xFFFF0F00) {
                case 256:
                  if ((this.nullCheckTypes[i] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)expression);
                    break;
                  } 
                  if ((this.nullCheckTypes[i] & 0x1000) == 0)
                    scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)expression); 
                  break;
                case 512:
                  if ((this.nullCheckTypes[i] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)expression);
                    break;
                  } 
                  scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)expression);
                  break;
                case 768:
                  scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)expression);
                  break;
                case 1024:
                  scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)expression);
                  break;
              } 
              break;
            } 
            if (flowInfo.isPotentiallyNull(local))
              switch (this.nullCheckTypes[i] & 0xFFFF0F00) {
                case 256:
                  this.nullReferences[i] = null;
                  if ((this.nullCheckTypes[i] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0)
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)expression); 
                  break;
                case 512:
                  this.nullReferences[i] = null;
                  if ((this.nullCheckTypes[i] & 0xFF & 0xFFFF0FFF) == 1 && (expression.implicitConversion & 0x400) != 0)
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)expression); 
                  break;
              }  
            break;
          case 3:
            if (flowInfo.isDefinitelyNull(local)) {
              scope.problemReporter().localVariableNullReference(local, location);
              break;
            } 
            if (flowInfo.isPotentiallyNull(local))
              scope.problemReporter().localVariablePotentialNullReference(local, location); 
            break;
          case 128:
            nullStatus = flowInfo.nullStatus(local);
            if (nullStatus != 4) {
              char[][] annotationName = scope.environment().getNonNullAnnotationName();
              scope.problemReporter().nullityMismatch((Expression)location, this.providedExpectedTypes[i][0], this.providedExpectedTypes[i][1], nullStatus, annotationName);
            } 
            break;
          case 16:
            checkUnboxing((Scope)scope, (Expression)location, flowInfo);
            break;
        } 
      } 
    } 
  }
  
  public String individualToString() {
    StringBuffer buffer = new StringBuffer("Finally flow context");
    buffer.append("[finalAssignments count - ").append(this.assignCount).append(']');
    buffer.append("[nullReferences count - ").append(this.nullCount).append(']');
    return buffer.toString();
  }
  
  public boolean isSubRoutine() {
    return true;
  }
  
  protected boolean recordFinalAssignment(VariableBinding binding, Reference finalAssignment) {
    if (this.assignCount == 0) {
      this.finalAssignments = new Reference[5];
      this.finalVariables = new VariableBinding[5];
    } else {
      if (this.assignCount == this.finalAssignments.length)
        System.arraycopy(
            this.finalAssignments, 
            0, 
            this.finalAssignments = new Reference[this.assignCount * 2], 
            0, 
            this.assignCount); 
      System.arraycopy(
          this.finalVariables, 
          0, 
          this.finalVariables = new VariableBinding[this.assignCount * 2], 
          0, 
          this.assignCount);
    } 
    this.finalAssignments[this.assignCount] = finalAssignment;
    this.finalVariables[this.assignCount++] = binding;
    return true;
  }
  
  public void recordUsingNullReference(Scope scope, LocalVariableBinding local, ASTNode location, int checkType, FlowInfo flowInfo) {
    if ((flowInfo.tagBits & 0x3) == 0 && !flowInfo.isDefinitelyUnknown(local)) {
      checkType |= this.tagBits & 0x1000;
      int checkTypeWithoutHideNullWarning = checkType & 0xFFFF0FFF;
      if ((this.tagBits & 0x1) != 0) {
        Expression reference;
        switch (checkTypeWithoutHideNullWarning) {
          case 256:
          case 257:
          case 512:
          case 513:
          case 769:
          case 1025:
            reference = (Expression)location;
            if (flowInfo.cannotBeNull(local)) {
              if (checkTypeWithoutHideNullWarning == 512) {
                if ((checkType & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNonNull(local, (ASTNode)reference); 
                flowInfo.initsWhenFalse().setReachMode(2);
              } else if (checkTypeWithoutHideNullWarning == 256) {
                scope.problemReporter().localVariableNonNullComparedToNull(local, (ASTNode)reference);
                flowInfo.initsWhenTrue().setReachMode(2);
              } 
              return;
            } 
            if (flowInfo.canOnlyBeNull(local)) {
              switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
                case 256:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  if ((checkType & 0x1000) == 0)
                    scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)reference); 
                  flowInfo.initsWhenFalse().setReachMode(2);
                  return;
                case 512:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)reference);
                  flowInfo.initsWhenTrue().setReachMode(2);
                  return;
                case 768:
                  scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)reference);
                  return;
                case 1024:
                  scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)reference);
                  return;
              } 
              break;
            } 
            if (flowInfo.isPotentiallyNull(local))
              switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
                case 256:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  break;
                case 512:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  break;
              }  
            break;
          case 3:
            if (flowInfo.cannotBeNull(local))
              return; 
            if (flowInfo.canOnlyBeNull(local)) {
              scope.problemReporter().localVariableNullReference(local, location);
              return;
            } 
            break;
        } 
      } else {
        Expression reference;
        switch (checkTypeWithoutHideNullWarning) {
          case 256:
          case 512:
            if (flowInfo.isDefinitelyNonNull(local)) {
              if (checkTypeWithoutHideNullWarning == 512) {
                if ((checkType & 0x1000) == 0)
                  scope.problemReporter().localVariableRedundantCheckOnNonNull(local, location); 
                flowInfo.initsWhenFalse().setReachMode(2);
              } else {
                scope.problemReporter().localVariableNonNullComparedToNull(local, location);
                flowInfo.initsWhenTrue().setReachMode(2);
              } 
              return;
            } 
          case 257:
          case 513:
          case 769:
          case 1025:
            reference = (Expression)location;
            if (flowInfo.isDefinitelyNull(local)) {
              switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
                case 256:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  if ((checkType & 0x1000) == 0)
                    scope.problemReporter().localVariableRedundantCheckOnNull(local, (ASTNode)reference); 
                  flowInfo.initsWhenFalse().setReachMode(2);
                  return;
                case 512:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariableNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  scope.problemReporter().localVariableNullComparedToNonNull(local, (ASTNode)reference);
                  flowInfo.initsWhenTrue().setReachMode(2);
                  return;
                case 768:
                  scope.problemReporter().localVariableRedundantNullAssignment(local, (ASTNode)reference);
                  return;
                case 1024:
                  scope.problemReporter().localVariableNullInstanceof(local, (ASTNode)reference);
                  return;
              } 
              break;
            } 
            if (flowInfo.isPotentiallyNull(local))
              switch (checkTypeWithoutHideNullWarning & 0xFFFF0F00) {
                case 256:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  break;
                case 512:
                  if ((checkTypeWithoutHideNullWarning & 0xFF) == 1 && (reference.implicitConversion & 0x400) != 0) {
                    scope.problemReporter().localVariablePotentialNullReference(local, (ASTNode)reference);
                    return;
                  } 
                  break;
              }  
            break;
          case 3:
            if (flowInfo.isDefinitelyNull(local)) {
              scope.problemReporter().localVariableNullReference(local, location);
              return;
            } 
            if (flowInfo.isPotentiallyNull(local)) {
              scope.problemReporter().localVariablePotentialNullReference(local, location);
              return;
            } 
            if (flowInfo.isDefinitelyNonNull(local))
              return; 
            break;
        } 
      } 
      recordNullReference(local, location, checkType, flowInfo);
    } 
  }
  
  void removeFinalAssignmentIfAny(Reference reference) {
    for (int i = 0; i < this.assignCount; i++) {
      if (this.finalAssignments[i] == reference) {
        this.finalAssignments[i] = null;
        this.finalVariables[i] = null;
        return;
      } 
    } 
  }
  
  protected void recordNullReference(LocalVariableBinding local, ASTNode expression, int checkType, FlowInfo nullInfo) {
    if (this.nullCount == 0) {
      this.nullLocals = new LocalVariableBinding[5];
      this.nullReferences = new ASTNode[5];
      this.nullCheckTypes = new int[5];
    } else if (this.nullCount == this.nullLocals.length) {
      int newLength = this.nullCount * 2;
      System.arraycopy(this.nullLocals, 0, 
          this.nullLocals = new LocalVariableBinding[newLength], 0, 
          this.nullCount);
      System.arraycopy(this.nullReferences, 0, 
          this.nullReferences = new ASTNode[newLength], 0, 
          this.nullCount);
      System.arraycopy(this.nullCheckTypes, 0, 
          this.nullCheckTypes = new int[newLength], 0, 
          this.nullCount);
    } 
    this.nullLocals[this.nullCount] = local;
    this.nullReferences[this.nullCount] = expression;
    this.nullCheckTypes[this.nullCount++] = checkType;
  }
  
  public void recordUnboxing(Scope scope, Expression expression, int nullStatus, FlowInfo flowInfo) {
    if (nullStatus == 2) {
      super.recordUnboxing(scope, expression, nullStatus, flowInfo);
    } else {
      recordNullReference(null, (ASTNode)expression, 16, flowInfo);
    } 
  }
  
  protected boolean internalRecordNullityMismatch(Expression expression, TypeBinding providedType, FlowInfo flowInfo, int nullStatus, TypeBinding expectedType, int checkType) {
    if (nullStatus == 1 || ((
      this.tagBits & 0x1) != 0 && nullStatus != 2)) {
      recordProvidedExpectedTypes(providedType, expectedType, this.nullCount);
      recordNullReference(expression.localVariableBinding(), (ASTNode)expression, checkType, flowInfo);
      return true;
    } 
    return false;
  }
}
